
#ifndef __libcoll_h__
#define __libcoll_h__

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/uio.h>

#ifdef __cplusplus
extern "C"
{
#endif

  /**
   * \brief LIBCOLL_IN_PLACE definition.
   */
#define LIBCOLL_IN_PLACE ((void *) -1L)

#define LIBCOLL_NODEADDR_NINTS      4

/*
 *  Adapter index: 1 - pami
 *                 2 - mpi
 */
extern int adapter_idx;
 
// Create enums for LIBCOLL_XFER_BROADCAST etc
// as well as an extern char *LIBCOLL_xfer_names[] array of strings.
#undef LIBCOLL_MYPROCESS
#undef LIBCOLL_MYPROCESS3
#define LIBCOLL_XFER_TYPE_LIST \
    LIBCOLL_MYPROCESS3(LIBCOLL_XFER_BROADCAST,     "BROADCAST", 0),     \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_ALLREDUCE,      "ALLREDUCE"),        \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_REDUCE,         "REDUCE"),           \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_ALLGATHER,      "ALLGATHER"),        \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_ALLGATHERV,     "ALLGATHERV"),       \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_ALLGATHERV_INT, "ALLGATHERV_INT"),   \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_SCATTER,        "SCATTER"),          \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_SCATTERV,       "SCATTERV"),         \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_SCATTERV_INT,   "SCATTERV_INT"),     \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_GATHER,         "GATHER"),           \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_GATHERV,        "GATHERV"),          \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_GATHERV_INT,    "GATHERV_INT"),      \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_BARRIER,        "BARRIER"),          \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_ALLTOALL,       "ALLTOALL"),         \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_ALLTOALLV,      "ALLTOALLV"),        \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_ALLTOALLV_INT,  "ALLTOALLV_INT"),    \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_SCAN,           "SCAN"),             \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_REDUCE_SCATTER, "REDUCE_SCATTER"),   \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_AMBROADCAST,    "AMBROADCAST"),      \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_AMSCATTER,      "AMSCATTER"),        \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_AMGATHER,       "AMGATHER"),         \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_AMREDUCE,       "AMREDUCE"),         \
    LIBCOLL_MYPROCESS(LIBCOLL_XFER_COUNT,          "N/A"),              \
    LIBCOLL_MYPROCESS3(LIBCOLL_XFER_TYPE_EXT,      "EXTENSIONS", 1000)
      /**< Begin extension-specific values */

#define LIBCOLL_MYPROCESS(e, s)     e
#define LIBCOLL_MYPROCESS3(e, s, n) e = n
  /**
   * \brief Collective operation transfer types; used by geometry and xfer routines
   */
  typedef enum
  {
    LIBCOLL_XFER_TYPE_LIST
  } libcoll_xfer_type_t;
#undef LIBCOLL_MYPROCESS
#undef LIBCOLL_MYPROCESS3
// This array of strings is only valid for libcoll_xfer_type_t entries
// LIBCOLL_XFER_BROADCAST through LIBCOLL_XFER_COUNT - 1.
extern const char *LIBCOLL_predefined_xfer_names[];

  /**
   * \brief LIBCOLL result status codes.
   */
  typedef enum
  {
    LIBCOLL_SUCCESS =  0, /**< Successful execution        */
    LIBCOLL_NERROR  = -1, /**< Generic error (-1)          */
    LIBCOLL_ERROR   =  1, /**< Generic error (+1)          */
    LIBCOLL_INVAL,        /**< Invalid argument            */
    LIBCOLL_UNIMPL,       /**< Function is not implemented */
    LIBCOLL_EAGAIN,       /**< Not currently available     */
    LIBCOLL_ENOMEM,       /**< Out of memory               */
    LIBCOLL_SHUTDOWN,     /**< Task has shutdown           */
    LIBCOLL_CHECK_ERRNO,  /**< Check the errno val         */
    LIBCOLL_PASSTHROUGH,  /**< tells the caller to use their own native call */
    LIBCOLL_OTHER,        /**< Other undefined error       */
    LIBCOLL_RESULT_EXT = 1000 /**< Begin extension-specific values */
  } libcoll_result_t;

  typedef enum
  {
    LIBCOLL_BYTE=0,        
    LIBCOLL_SIGNED_CHAR,
    LIBCOLL_SIGNED_SHORT,
    LIBCOLL_SIGNED_INT,
    LIBCOLL_SIGNED_LONG,
    LIBCOLL_SIGNED_LONG_LONG,    
    LIBCOLL_UNSIGNED_CHAR,
    LIBCOLL_UNSIGNED_SHORT,
    LIBCOLL_UNSIGNED_INT,
    LIBCOLL_UNSIGNED_LONG,
    LIBCOLL_UNSIGNED_LONG_LONG,    
    LIBCOLL_FLOAT,
    LIBCOLL_DOUBLE,
    LIBCOLL_LONG_DOUBLE,    
    LIBCOLL_SINGLE_COMPLEX,
    LIBCOLL_DOUBLE_COMPLEX,    
    LIBCOLL_LOC_2INT,
    LIBCOLL_LOC_2FLOAT,
    LIBCOLL_LOC_2DOUBLE,
    LIBCOLL_LOC_SHORT_INT,
    LIBCOLL_LOC_FLOAT_INT,
    LIBCOLL_LOC_DOUBLE_INT,
    LIBCOLL_LOC_LONG_INT,
    LIBCOLL_LOC_LONGDOUBLE_INT,    
    LIBCOLL_TYPE_COUNT,
    LIBCOLL_TYPE_UNDEFINED,
  } libcoll_type_t;
  
  typedef enum
  {
    LIBCOLL_MAX=0, 
    LIBCOLL_MIN, 
    LIBCOLL_SUM, 
    LIBCOLL_PROD, 
    LIBCOLL_LAND, 
    LIBCOLL_LOR, 
    LIBCOLL_LXOR, 
    LIBCOLL_BAND, 
    LIBCOLL_BOR, 
    LIBCOLL_BXOR, 
    LIBCOLL_MAXLOC, 
    LIBCOLL_MINLOC,     
    LIBCOLL_OP_COUNT,
  } libcoll_eop_t;
  
#define LIBCOLL_OP_UNDEFINED  (unsigned) -1
  typedef unsigned libcoll_op_t;

  typedef void*    libcoll_context_t;    /**< \ingroup context Context for data transfers       */
  typedef void*    libcoll_epdomain_t;   /**< The Endpoint Domain handle */
  typedef uint32_t libcoll_task_t;       /**< Identity of a process (should we depricate this, subsumed by endpoint??)            */
  typedef uint32_t libcoll_endpoint_t;   /**< Endpoint of a p2p operation      */
  typedef void*    libcoll_geometry_t;   /**< Geometry object handle           */
  typedef void*    libcoll_algorithm_t;  /**< Algorithm object handle          */

  /**
   *  This enum contains the "reason" codes that indicate why a
   *  particular collective operation could fail.  Upon querying the
   *  metadata of a particular collective the user may have a function
   *  pointer populated in the "check_fn" field of the metadata
   *  struct.  The user can pass in the call site parameters (the
   *  libcoll_xfer_t) of the collective into this function and receive a
   *  list of failures.
   */
  typedef union 
  {
    unsigned bitmask;                 /**< Raw bitmask data              */
    struct
    {
      unsigned unspecified       : 1; /**< Unspecified failure              */
      unsigned range             : 1; /**< Send/Recv bytes are out of range */
      unsigned datatype_op       : 1; /**< Datatype/operation not valid     */
      unsigned contiguous_send   : 1; /**< Send data must be contiguous     */
      unsigned contiguous_recv   : 1; /**< Receive data must be contiguous  */
    } check;                          /**< Metadata 'reason' information     */
  } libcoll_metadata_result_t;

  struct libcoll_xfer_t;
  /**
   * \brief Function signature for metadata queries
   *
   * \param [in] in A "call-site"   collective query parameters
   * \retval libcoll_metadata_check  failure code, 0=success, nonzero
   *                                 bits are OR'd from libcoll_metadata_check_t
   */
  typedef libcoll_metadata_result_t (*libcoll_metadata_function) (struct libcoll_xfer_t *in);


  /**
   * \brief A metadata structure to describe a collective protocol
   *
   * \b Correctness \b Parameters
   *
   * If an algorithm is placed on the "must query list", then the user
   * must use the metadata to determine if the protocol can be issued
   * given the call site parameters of the collective operation.
   *
   * This may include calling the check_fn function, which takes as
   * input a libcoll_xfer_t corresponding to the call site parameters.
   * The call site is defined as the code location where the user will
   * call the LibColl_Start function.
   *
   * The semantics for when the user must call the check_fn function
   * are as follows: If check_fn is NULL, and checkrequired is 0, the
   * user cannot and is not required to call the check_fn function.
   * The user is still required to check the other bits in the
   * bitmask_correct to determine if the collective parameters satisfy
   * the requirements.
   *
   * check_fn will never be NULL when checkrequired is set to 1
   *
   * When check_fn is non-NULL and checkrequired is 1, the user must
   * call the check_fn function for each call to LIBCOLL_Start to
   * determine if the collective is valid.  check_fn will return local
   * validity of the parameters, and no other bits in the struct need
   * to be checked. If the nonlocal bit is set in the return code, a
   * reduction will be required to determine the validity of the
   * protocol across tasks.
   *
   */
  typedef struct
  {
    char                 * name;       /**< A character string
					  describing the protocol  */
    unsigned               version;    /**< A version number for the
					  protocol  */
    libcoll_metadata_function check_fn;   /**<  A function pointer to validate parameters for the collective operation.  Can be NULL if no correctness check is required */
    size_t                 range_lo;    /**<  This protocol has minimum bytes to send/recv requirements */
    size_t                 range_hi;    /**<  This protocol has maximum bytes to send/recv requirements  */
    /**
     * \brief Bitmask to determine if certain checks are required.
     */
    union
    {
      unsigned             bitmask_correct;  /**<  Raw bitmask data   */
      struct
      {
        unsigned           checkrequired: 1; /**<  A flag indicating whether or not the user MUST query the metadata or not.  0:query is not necessary.  1: the user must call the check_fn to determine if the call site parameters are correct */
        unsigned           rangeminmax  : 1; /**<  This protocol only supports a range of bytes sent/received. 0: no min/max, 1: check range_lo/range_hi                   */
        unsigned           blocking     : 1; /**<  This protocol may block.
						0: not blocking,
						1: may block  */
        unsigned           inplace      : 1; /**<  This protocol supports in-place buffering:
						0: the send and recv buffers my not overlap,
						1: the send and recv buffers may overlap     */
        unsigned           asyncflowctl : 1; /**<  This protocol is asynchronous and may require
						some flow control (barrier) to control
						excessive unexpected data buffering.         */
        unsigned           oneatatime   : 1; /**<  This protocol only permits one collective
						to be out-standing at a time                 */
      } values;
    } check_correct;
    
    /**
     * \brief Performance parameters
     */
    union
    {
      unsigned             bitmask_perf;     /**<  Raw bitmask data  */
      struct
      {
        unsigned           hw_accel     : 1; /**<  This collective is using special purpose hardware to accelerate the collective 0:  no 1:  yes */
      } values;
    } check_perf;
  } libcoll_metadata_t;



  typedef void (*libcoll_event_function) ( libcoll_context_t   context,
					   void           * cookie,
					   libcoll_result_t    result );


  /**
   * \brief enable operation b[i] = a[i] op b[i]. We only support
   * commutative operations
   */
  typedef void (*libcoll_math_function) ( void    *b,
					  void    *a,
					  int      nelem, 
					  libcoll_type_t  dt );    

  libcoll_op_t  LIBCOLL_Register_math_function ( libcoll_context_t   ctxt,
						 libcoll_math_function mfn );
  
  libcoll_result_t  LIBCOLL_Unregister_math_function (libcoll_context_t  ctxt,
						      libcoll_op_t       op);
  
  /**
   * \brief Non-blocking barrier operation parameters.
   *
   * \note The barrier collective operation does not require any additional
   *       parameters.  This structure is defined for completeness purposes.
   */
  typedef struct
  {
      int dummy; /* Silence compiler warning...*/
  } libcoll_barrier_t;

  /**
   * \brief Non-blocking broadcast operation parameters
   */
  typedef struct
  {
    libcoll_endpoint_t   root;       /**< The endpoint performing the broadcast */
    char               * buf;       /**< Source buffer to broadcast on root, destination buffer on non-root */
    libcoll_type_t       type;       /**< Data type layout; may be different on root/destinations */
    size_t               typecount; /**< Single type replication count */
  } libcoll_broadcast_t;

  /**
   * \brief Non-blocking allreduce operation parameters.
   */
  typedef struct
  {
    char               * sndbuf;    /**< Source buffer */
    libcoll_type_t       stype;     /**< Source buffer type; ignored if sndbuf == LIBCOLL_IN_PLACE */
    size_t               stypecount; /**< Source buffer type count; ignored if sndbuf == LIBCOLL_IN_PLACE */
    char               * rcvbuf;    /**< Receive buffer */
    libcoll_type_t       rtype;     /**< Receive buffer type */
    size_t               rtypecount;/**< Receive buffer type count */
    libcoll_op_t         op;        /**< Reduce operation */
  } libcoll_allreduce_t;


  /**
   * \brief Non-blocking alltoall vector operation parameters.
   */
  typedef struct
  {
    char             * sndbuf;        /**< The base address of the buffers containing data to be sent */
    libcoll_type_t     stype;         /**< Send buffer type; ignored if sndbuf == libcoll_IN_PLACE */
    size_t           * stypecounts;   /**< Array of type replication counts; size of geometry length;
                                           ignored if sndbuf == libcoll_IN_PLACE */
    size_t           * sdispls;       /**< Array of stype counts into the sndbuf; size of geometry length;
                                           ignored if sndbuf == libcoll_IN_PLACE */
    char             * rcvbuf;        /**< The base address of the buffer for data reception */
    libcoll_type_t     rtype;         /**< Receive buffer type */
    size_t           * rtypecounts;   /**< Array of type replication counts; size of geometry length */
    size_t           * rdispls;       /**< Array of rtype counts into the rcvbuf; size of geometry length */
  } libcoll_alltoallv_t;

  /**
   * \brief Non-blocking alltoall vector operation parameters.
   */
  typedef struct
  {
    char              * sndbuf;      /**< The base address of the buffers containing data to be sent */
    libcoll_type_t      stype;       /**< Send buffer type; ignored if sndbuf == libcoll_IN_PLACE */
    int               * stypecounts; /**< Array of type replication counts; size of geometry length;
                                          ignored if sndbuf == libcoll_IN_PLACE */
    int               * sdispls;     /**< Array of stype counts into the sndbuf; size of geometry length;
                                          ignored if sndbuf == libcoll_IN_PLACE */
    char              * rcvbuf;      /**< The base address of the buffer for data reception */
    libcoll_type_t      rtype;       /**< Receive buffer type */
    int               * rtypecounts; /**< Array of type replication counts; size of geometry length */
    int               * rdispls;     /**< Array of rtype counts into the rcvbuf; size of geometry length */
  } libcoll_alltoallv_int_t;

  /**
   * \brief Non-blocking alltoall operation parameters.
   */
  typedef struct
  {
    char        * sndbuf;     /**< The base address of the buffers containing data to be sent */
    libcoll_type_t   stype;      /**< Send buffer type; ignored if sndbuf == libcoll_IN_PLACE */
    size_t        stypecount; /**< Send buffer type count; ignored if sndbuf == libcoll_IN_PLACE */
    char        * rcvbuf;     /**< The base address of the buffer for data reception */
    libcoll_type_t   rtype;      /**< Receive buffer type */
    size_t        rtypecount; /**< Receive buffer type count */
  } libcoll_alltoall_t;

  /**
   * \brief Non-blocking reduce operation parameters
   */
  typedef struct
  {
    libcoll_endpoint_t      root;        /**< The endpoint of the reduction root */
    char               * sndbuf;      /**< Send buffer */
    libcoll_type_t          stype;       /**< Send buffer type; ignored if sndbuf == libcoll_IN_PLACE */
    size_t               stypecount;  /**< Send buffer type count; ignored if sndbuf == libcoll_IN_PLACE */
    char               * rcvbuf;      /**< Receive buffer */
    libcoll_type_t          rtype;       /**< Receive buffer type */
    size_t               rtypecount;  /**< Receive buffer type count */
    libcoll_op_t   op;          /**< Reduce operation */
  } libcoll_reduce_t;

  /**
   * \brief Non-blocking reduce_scatter operation parameters
   */
  typedef struct
  {
    char               * sndbuf;      /**< Send buffer */
    libcoll_type_t          stype;       /**< Send buffer type */
    size_t               stypecount;  /**< Send buffer type count */
    char               * rcvbuf;      /**< Receive buffer */
    libcoll_type_t          rtype;       /**< Receive buffer type */
    size_t             * rcounts;     /**< Number of elements to receive from the destination,
                                           common on all nodes; index of my task is the count
                                           my task will receive */
    libcoll_op_t   op;          /**< Reduce operation */
    // void               * data_cookie; /**< Cookie to provide to each individual reduce operation */
    // int                  commutative; /**< Inputs to operation are commutative */
  } libcoll_reduce_scatter_t;


  /**
   * \brief Non-blocking allgather parameters
   */
  typedef struct
  {
    char        * sndbuf;     /**< Send buffer */
    libcoll_type_t   stype;      /**< Data type of send buffer; ignored if sndbuf == libcoll_IN_PLACE */
    size_t        stypecount; /**< Send type replication count; ignored if sndbuf == libcoll_IN_PLACE */
    char        * rcvbuf;     /**< Receive buffer */
    libcoll_type_t   rtype;      /**< Data type of receive buffer */
    size_t        rtypecount; /**< Receive type replication count */
  } libcoll_allgather_t;


  /**
   * \brief Non-blocking gather parameters
   */
  typedef struct
  {
    libcoll_endpoint_t    root;       /**< The gather root endpoint */
    char                * sndbuf;     /**< Send buffer */
    libcoll_type_t        stype;      /**< Data type of send buffer; ignored if sndbuf == libcoll_IN_PLACE */
    size_t                stypecount; /**< Send type replication count; ignored if sndbuf == libcoll_IN_PLACE */
    char                * rcvbuf;     /**< Receive buffer */
    libcoll_type_t        rtype;      /**< Data type of receive buffer */
    size_t                rtypecount; /**< Receive type replication count */
  } libcoll_gather_t;

  /**
   * \brief Non-blocking gatherv with size_t-type arrays parameters
   */
  typedef struct
  {
    libcoll_endpoint_t   root;        /**< The gatherv root endpoint */
    char            * sndbuf;      /**< Base address of the buffers containing data to be sent */
    libcoll_type_t       stype;       /**< Type of all source buffers; ignored if sndbuf == libcoll_IN_PLACE */
    size_t            stypecount;  /**< Source type replication count; ignored if sndbuf == libcoll_IN_PLACE */
    char            * rcvbuf;      /**< Receive buffer */
    libcoll_type_t       rtype;       /**< Receive type */
    size_t          * rtypecounts; /**< Array of type replication counts; size of geometry length */
    size_t          * rdispls;     /**< Array of rtype counts into the rcvbuf; size of geometry length */
  } libcoll_gatherv_t;

  /**
   * \brief Non-blocking gatherv with integer-type arrays parameters
   */
  typedef struct
  {
    libcoll_endpoint_t   root;        /**< The gatherv_int root node endpoint */
    char            * sndbuf;      /**< Base address of the buffers containing data to be sent */
    libcoll_type_t       stype;       /**< Type of all source buffers; ignored if sndbuf == libcoll_IN_PLACE */
    int               stypecount;  /**< Source type replication count; ignored if sndbuf == libcoll_IN_PLACE */
    char            * rcvbuf;      /**< Receive buffer */
    libcoll_type_t       rtype;       /**< Receive type */
    int             * rtypecounts; /**< Array of type replication counts; size of geometry length */
    int             * rdispls;     /**< Array of rtype counts into the rcvbuf; size of geometry length */
  } libcoll_gatherv_int_t;

  /**
   * \brief Non-blocking allgatherv with size_t-type arrays parameters
   */
  typedef struct
  {
    char        * sndbuf;      /**< Base address of the buffers containing data to be sent */
    libcoll_type_t   stype;       /**< Type of all source buffers; ignored if sndbuf == libcoll_IN_PLACE */
    size_t        stypecount;  /**< Source type replication count; ignored if sndbuf == libcoll_IN_PLACE */
    char        * rcvbuf;      /**< Receive buffer */
    libcoll_type_t   rtype;       /**< Receive type */
    size_t      * rtypecounts; /**< Array of type replication counts; size of geometry length */
    size_t      * rdispls;     /**< Array of rtype counts into the rcvbuf; size of geometry length */
  } libcoll_allgatherv_t;

  /**
   * \brief Non-blocking allgatherv with integer-type arrays parameters.
   */
  typedef struct
  {
    char        * sndbuf;      /**< Base address of the buffers containing data to be sent */
    libcoll_type_t   stype;       /**< Type of all source buffers; ignored if sndbuf == libcoll_IN_PLACE */
    int           stypecount;  /**< Source type replication count; ignored if sndbuf == libcoll_IN_PLACE */
    char        * rcvbuf;      /**< Receive buffer */
    libcoll_type_t   rtype;       /**< Receive type */
    int         * rtypecounts; /**< Array of type replication counts; size of geometry length */
    int         * rdispls;     /**< Array of rtype counts into the rcvbuf; size of geometry length */
  } libcoll_allgatherv_int_t;


  /**
   * \brief Non-blocking scatter operation parameters.
   */
  typedef struct
  {
    libcoll_endpoint_t   root;       /**< The scatter root node endpoint */
    char            * sndbuf;     /**< Source buffer */
    libcoll_type_t       stype;      /**< Source buffer type */
    size_t            stypecount; /**< Source buffer type count */
    char            * rcvbuf;     /**< Receive buffer */
    libcoll_type_t       rtype;      /**< Receive buffer type; ignored if rcvbuf == libcoll_IN_PLACE */
    size_t            rtypecount; /**< Receive buffer type count; ignored if rcvbuf == libcoll_IN_PLACE */
  } libcoll_scatter_t;

  /**
   * \brief Non-blocking scatterv operation parameters.
   */
  typedef struct
  {
    libcoll_endpoint_t   root;        /**< The scatterv root node endpoint */
    char            * sndbuf;      /**< Base address of the buffers containing data to be sent */
    libcoll_type_t       stype;       /**< Source type for all buffers */
    size_t          * stypecounts; /**< Array of type replication counts; size of geometry length */
    size_t          * sdispls;     /**< Array of stype counts into the sndbuf; size of geometry length */
    char            * rcvbuf;      /**< Receive buffer */
    libcoll_type_t       rtype;       /**< Receive buffer type; ignored if rcvbuf == libcoll_IN_PLACE */
    size_t            rtypecount;  /**< Receive buffer type count; ignored if rcvbuf == libcoll_IN_PLACE */
  } libcoll_scatterv_t;

  /**
   * \brief Non-blocking scatterv_int operation parameters.
   */
  typedef struct
  {
    libcoll_endpoint_t   root;        /**< The scatterv_int root node endpoint */
    char            * sndbuf;      /**< Base address of the buffers containing data to be sent */
    libcoll_type_t       stype;       /**< Source type for all buffers */
    int             * stypecounts; /**< Array of type replication counts; size of geometry length */
    int             * sdispls;     /**< Array of stype counts into the sndbuf; size of geometry length */
    char            * rcvbuf;      /**< Receive buffer */
    libcoll_type_t       rtype;       /**< Receive buffer type; ignored if rcvbuf == libcoll_IN_PLACE */
    int               rtypecount;  /**< Receive buffer type count; ignored if rcvbuf == libcoll_IN_PLACE */
  } libcoll_scatterv_int_t;

  /**
   * \brief Non-blocking scan operation parameters.
   */
  typedef struct
  {
    char               * sndbuf;      /**< Source buffer */
    libcoll_type_t          stype;       /**< Source buffer type; ignored if sndbuf == libcoll_IN_PLACE */
    size_t               stypecount;  /**< Source buffer type count; ignored if sndbuf == libcoll_IN_PLACE */
    char               * rcvbuf;      /**< Receive buffer */
    libcoll_type_t          rtype;       /**< Receive buffer type */
    size_t               rtypecount;  /**< Receive buffer type count */
    libcoll_op_t   op;          /**< Reduce operation */
    //void               * data_cookie; /**< Cookie to provide to each individual reduce operation */
    int                  exclusive;   /**< Scan operation is exclusive of current node */
  } libcoll_scan_t;


  typedef union
  {
    libcoll_barrier_t          xfer_barrier;
    libcoll_broadcast_t        xfer_broadcast;
    libcoll_allreduce_t        xfer_allreduce;
    libcoll_reduce_t           xfer_reduce;         /**< Reduce collective parameters */
    libcoll_allgather_t        xfer_allgather;      /**< Allgather collective parameters */
    libcoll_allgatherv_t       xfer_allgatherv;     /**< Allgatherv collective parameters */
    libcoll_allgatherv_int_t   xfer_allgatherv_int; /**< Allgatherv_int collective parameters */
    libcoll_scatter_t          xfer_scatter;        /**< Scatter collective parameters */
    libcoll_scatterv_t         xfer_scatterv;       /**< Scatterv collective parameters */
    libcoll_scatterv_int_t     xfer_scatterv_int;   /**< Scatterv_int collective parameters */
    libcoll_gather_t           xfer_gather;         /**< Gather collective parameters */
    libcoll_gatherv_t          xfer_gatherv;        /**< Gatherv collective parameters */
    libcoll_gatherv_int_t      xfer_gatherv_int;    /**< Gatherv_int collective parameters */
    libcoll_alltoall_t         xfer_alltoall;       /**< Alltoall collective parameters */
    libcoll_alltoallv_t        xfer_alltoallv;      /**< Alltoallv collective parameters */
    libcoll_alltoallv_int_t    xfer_alltoallv_int;  /**< Alltoallv_int collective parameters */
    libcoll_scan_t             xfer_scan;           /**< Scan collective parameters */
    libcoll_reduce_scatter_t   xfer_reduce_scatter; /**< Reduce-Scatter collective parameters */
  } libcoll_collective_t;

  /**
   * \brief Collective operation parameter structure
   */
  typedef struct libcoll_xfer_t
  {
    libcoll_event_function      cb_done;   /**< Callback to invoke on collective completion */
    void                      * cookie;    /**< Cookie to deliver to cb_done function */
    libcoll_algorithm_t         algorithm; /**< Algorithm object from LIBCOLL_Geometry_algorithms_query() */
    libcoll_geometry_t          geometry;  /**< Pass in the geometry pointer */
    libcoll_collective_t        cmd;       /**< Collective specific arguments */
  } libcoll_xfer_t;
  
  /**
   * \brief Start a collective on a context
   *
   * For each endpoint specified in the geometry, a call to LIBCOLL_Collective must
   * be made to the corresponding context.
   *
   * It is illegal to post a collective via LIBCOLL_Collective() to a context
   * that is not accociated with an endpoint in the geometry
   *
   * \param [in] context     context to deliver async callback
   * \param [in] cmd         Collective operation arguments
   *
   * \retval ::LIBCOLL_SUCCESS  The collective operation has started successfully.
   */
  libcoll_result_t LIBCOLL_Start (libcoll_context_t context, libcoll_xfer_t * cmd);

  typedef enum
  {
    LIBCOLL_SELECTION_FAILSAFE = 0, /// Choose a failsafe algorthm

    LIBCOLL_SELECTION_RR = 1,       /// Cycle over all algorithms

    LIBCOLL_SELECTION_PERF_STATIC,  /// Choose best algorithm via table lookup

    LIBCOLL_SELECTION_PERF_DYNAMIC, /// Measure performance of optimized
                                    /// algorithms for a given
                                    /// callsite. Dynamically select
                                    /// the best algorithm
    LIBCOLL_SELECTION_COUNT
  } libcoll_collselect_type_t;

  /**
   * \brief Execute collective via collective selection in LIBCOLL.
   * The algorithm passed in will be ignored.
   */
  libcoll_result_t LIBCOLL_AutoSelect (libcoll_context_t           context, 
				       libcoll_collselect_type_t   cstype, 
				       libcoll_xfer_type_t         xfer_type,
				       libcoll_xfer_t            * cmd);

  libcoll_result_t LIBCOLL_AutoSelect_no_defer (libcoll_context_t           context,
               libcoll_collselect_type_t   cstype,
               libcoll_xfer_type_t         xfer_type,
               libcoll_xfer_t            * cmd);

  libcoll_result_t temporary_function_add_algorithms_to_geometry(
    libcoll_geometry_t geometry);

  libcoll_result_t LIBCOLL_Geometry_algorithms_num(libcoll_geometry_t geometry,
						 libcoll_xfer_type_t coll_type,
						 size_t       *lists_lengths);
  
  libcoll_result_t LIBCOLL_Geometry_algorithms_query(libcoll_geometry_t geometry,
						     libcoll_xfer_type_t  colltype,
						     libcoll_algorithm_t *algs0,
						     libcoll_metadata_t  *mdata0,
						     size_t             num0,
						     libcoll_algorithm_t *algs1,
						     libcoll_metadata_t  *mdata1,
						     size_t             num1);

  libcoll_result_t LIBCOLL_Advance (libcoll_context_t context);
  libcoll_result_t LIBCOLL_Advance_pami (libcoll_context_t context);
  libcoll_result_t LIBCOLL_Advance_mpi (libcoll_context_t context);

  ///
  /// \brief Opaque libcoll node address for each task, typically used
  /// in hybrid collectives. The caller of libcoll can store the node
  /// address for all the endpoints in this task.
  ///
  typedef struct {
    uint32_t  node_addr[LIBCOLL_NODEADDR_NINTS];  //Task node address
  } libcoll_nodeaddr_t;
  
  ///
  /// \brief the libcoll endpoint address, that specifies the task
  /// node address and the offset in that task
  ///
  typedef struct {
    libcoll_nodeaddr_t   task;     //Node address of the task
    uint32_t             offset;   //Endpoint offset in the task
  } libcoll_epaddr_t;
  
  ///
  /// \brief the libcoll topology info callback function.	
  /// \param [in] epd_clientdata client data passed EPDomain
  /// \param [in] endpoint
  /// \param [out] epaddr   address structure filled in
  /// \param [out] numTasks number of tasks in this endpoint domain
  /// \param [out] numNodes number of SMP nodes in this endpoint domain
  ///
  typedef libcoll_result_t (*libcoll_ep_topology_query_fn_t) (void               * clientdata,
  							     libcoll_endpoint_t   endpoint,
							     libcoll_epaddr_t   * epaddr);
  
  ///
  /// \brief the libcoll peers query function. This function returns
  /// the peers of this endpoint on this SMP node. This calls is mainly used
  /// by hybrid collectives.	
  /// \brief [in] clientdata of the libcoll epdomain
  /// \brief [in] peers vector of peers for this endpoint filled in this call
  /// \brief [in] npeers_in size of the peers vector
  /// \brief [out] npeers_out actual number of peers
  ///
  typedef libcoll_result_t (*libcoll_ep_peers_query_fn_t)  (void                * clientdata,
							   libcoll_endpoint_t  * peers,	   
							   size_t                npeers_in,
							   size_t              * npeers_out);

  typedef struct {
    libcoll_ep_topology_query_fn_t    topology_query_fn;
    libcoll_ep_peers_query_fn_t       peers_query_fn;
  } libcoll_epdomain_callbacks_t;
  

  typedef enum {
    LIBCOLL_PAMI_CLIENT = 0,          //PAMI Client handle
    LIBCOLL_PAMI_CONTEXT_POINTER = 1, //Pointer to vector of PAMI context handles
    LIBCOLL_PAMI_NUM_CONTEXTS,        //Number of PAMI contexts to use
    LIBCOLL_JOB_ID,
    LIBCOLL_EXTERNAL_COMMUNICATOR_POINTER,
    LIBCOLL_NUM_ATTRIBUTES,
  } libcoll_attribute_id_t;

  typedef struct {
    libcoll_attribute_id_t    attribute_id;       //ID of the attribute
    void                    * attribute_value;    //Attribute value
  } libcoll_attribute_t;

  /**
   *  \brief Define a namespace of endpoints from
   *  0...num_eps-1. Location query functions are passed in by the
   *  caller to define where these endpoints are located. An EPDomain
   *  must be created before the geometry create call. However every
   *  geometry create call can share the EPDomain handles if the
   *  geometry endpoints are subsets of the EPDomain.
   *  
   *  LIBCOLL_EPDomain_create must be called once for each instance of
   *  an endpoint in a process.
   * 
   *  \param[in]  context  libcoll context
   *  \param[out] epdomain libcoll epdomain handle
   *  \param[in]  clientdata epdomain clientdata that is passed into the query callback fns
   *  \param[in]  attributes vector of attributes to be set
   *  \param[in]  nattributes number of attributes
   *  \param[in]  num_eps  number of endpoins in the EPDomain
   *  \param[in]  my_ep    caller's endpoint in the EPDomain
   *  \param[in]  cb       epdomain callback structure
   *  \retval  LIBCOLL_SUCCESS on success of the EPDomain creation
   */ 
  libcoll_result_t    LIBCOLL_EPDomain_create (libcoll_context_t            * context,
					       libcoll_epdomain_t           * epdomain,
					       void                         * clientdata, //????
					       libcoll_attribute_t          * attributes,
					       size_t                         nattributes,	
					       size_t                         num_eps,
					       libcoll_endpoint_t             my_ep,
					       libcoll_epdomain_callbacks_t   cb);
  
  libcoll_result_t    LIBCOLL_EPDomain_query_callbacks (libcoll_epdomain_t              epdomain,
							libcoll_epdomain_callbacks_t  * cb);
  
  /*
   * Create a geometry from a range of endpoints in an EPDomain. 
   *
   * \param [in]  context         The LibColl context handle. Callbacks pass this handle
   * \param [in]  epdomain        Endpoint Domain handle for this geometry
   * \param [in]  attributes      List of configurable attributes and values
   * \param [in]  nattributes     The number of attribute elements
   * \param [out] geometry        Opaque geometry object to initialize
   * \param [in]  parent          Parent geometry containing all the entries in the eplist. 
   * \param [in]  id              Identifier for this geometry(must be > 0)
   *                              which uniquely represents this geometry(if tasks overlap)
   * \param [in]  start_ep        start endpoint
   * \param [in]  num_eps         Number of endpoints
   * \param [in]  fn              event function to call when geometry has been created
   * \param [in]  cookie          user cookie to deliver with the callback
   *
   * \retval LIBCOLL_SUCCESS      The eprange geometry object initialization was started
   */ 
  libcoll_result_t    LIBCOLL_Geometry_create_eprange (libcoll_context_t        context,
						       libcoll_epdomain_t       epdomain,
						       libcoll_attribute_t    * attributes,
						       size_t                   nattributes,
						       libcoll_geometry_t     * geometry,
						       libcoll_geometry_t       parent,
						       size_t                   id,
						       libcoll_endpoint_t       start_ep,
						       size_t                   num_eps,
						       libcoll_event_function         fn,
						       void                   * clientdata);
    
  libcoll_result_t    LIBCOLL_Geometry_create_eplist (libcoll_context_t        context,
						      libcoll_epdomain_t       epdomain,
						      libcoll_attribute_t    * attributes,
						      size_t                   nattributes,
						      libcoll_geometry_t     * geometry,
						      libcoll_geometry_t       parent,
						      size_t                   id,
						      libcoll_endpoint_t     * eplist,
						      size_t                   num_eps,
						      libcoll_event_function         fn,
						      void                   * clientdata); 

  libcoll_result_t LIBCOLL_Geometry_destroy (libcoll_context_t          ctx,
                                             libcoll_geometry_t       * geometry,           //Geometry ptr
                                             libcoll_event_function     fn,          //completion callback
                                             void                     * cookie);

  /**
   *  \brief Initialize a libcoll context
   * 
   *  \param [in] name of the LIBCOLL Context
   *  \param [in] attributes for the LIBCOLL Context
   *  \param [in] number of attributes
   *  \param [in] completion callback
   *  \param [in] completion clientdata pointer
   *  \param [inout] context handle pointer
   *  \param [inout] pointer to world endpoint domain spanning 
   *                 all processes in the job.
   *  \param [inout] pointer to world geometry spanning all processes in the job.
   *  \retval LIBCOLL_SUCCESS or LIBCOLL_ERROR 
   *
   *  On successful completion of LIBCOLL_Init, LIBCOLL_advance() must
   *  be called to complete the non-blocking initialization call.
   *
   *  Note, when initializing the PAMI adapter the PAMI client, context
   *  handles and number of contexts must be passed in as attributes.
   *
   */
  libcoll_result_t LIBCOLL_Init(char                   * name,
				libcoll_attribute_t    * attributes,
				size_t                   nattr,
				libcoll_event_function   fn,
				void                   * clientdata,
				libcoll_context_t      * context,
				libcoll_epdomain_t     * world_epdomain,
				libcoll_geometry_t     * world_geometry);

libcoll_result_t LIBCOLL_Attach_EPD_Callbacks(
  size_t mpicomm, int key, size_t mpiparentcomm);
libcoll_result_t LIBCOLL_Get_EPD_Callbacks(size_t mpicomm,
  void **epd_callback_class_ptr);
libcoll_result_t LIBCOLL_Detach_EPD_Callbacks(size_t mpicomm);
libcoll_result_t LIBCOLL_topology_query_fn(void *clientdata,
  libcoll_endpoint_t endpoint, libcoll_epaddr_t *epaddr);
libcoll_result_t LIBCOLL_peers_query_fn(void *clientdata,
  libcoll_endpoint_t *peers, size_t npeers_in, size_t *npeers_out);

#ifdef __cplusplus
};
#endif

#endif
