/* Licensed Materials - Property of IBM                             */
/*                                                                  */
/* (C) Copyright IBM Corp. 2011 - 2017 All Rights Reserved           */
/* US Government Users Restricted Rights -                          */
/* Use, duplication, or disclosure restricted                       */
/* by GSA ADP Schedule Contract with IBM Corp.                      */
/*                                                                  */

/**
 * \file sys/lapiunix/pami_sys.h
 * \brief "Default" platform specifc type definitions for IBM's PAMI message layer.
 */

#ifndef __default_pami_sys_pre_h__
#define __default_pami_sys_pre_h__

#ifndef __64BIT__
#define __64BIT__
#endif

/* We need 64 bytes memory region for both 32- and 64-bit on PE implementation */
//#define PAMI_CLIENT_MEMREGION_SIZE_STATIC (8*sizeof(uintptr_t))
#define CUDAIPCMEMHANDLE_SIZE 64
#define PAMI_CLIENT_MEMREGION_SIZE_STATIC (16*sizeof(uintptr_t)+CUDAIPCMEMHANDLE_SIZE)
#define PAMI_WORK_SIZE_STATIC (8*sizeof(uintptr_t))
#define PAMI_RECV_REQUEST_SIZE_STATIC (8*sizeof(uintptr_t))

#else
#ifndef __default_pami_sys_post_h__
#define __default_pami_sys_post_h__

#define PAMI_Endpoint_create(__client,__task,__offset,__endpoint) ({ *(__endpoint) = (((__offset) << 23) | (__task)); PAMI_SUCCESS; })
#define PAMI_Endpoint_query(__endpoint,__task,__offset) ({ *(__task) = (__endpoint) & 0x007fffff; *(__offset) = ((__endpoint) >> 23) & 0x03f; PAMI_SUCCESS; })

#include "infiniband/verbs.h"

#ifdef __cplusplus
extern "C"
{
#endif

  /** PAMI IBV Internal Data Structures */
  typedef struct _pami_ibv_msgstate_t {
    uint32_t  freespace;
    uint32_t  untracked;
  } pami_ibv_messagestate_t;

  typedef struct _pami_ibv_qpstate_t {
    struct ibv_qp           ** qp;
    pami_ibv_messagestate_t  * msg_state;
    unsigned                   num_qp;
    unsigned                   max_inline_bytes;
  } pami_ibv_qpstate_t;

#define PAMI_IBV_MAX_METADATA_SIZE 14

  typedef struct _pami_ibv_PacketHeader_t {
    char               _metadata[PAMI_IBV_MAX_METADATA_SIZE]; //metadata
    unsigned short     _dispatchid;
  } pami_ibv_packetheader_t;

  typedef int (*pami_recvfunc_t) (pami_ibv_packetheader_t   * hdr,
                                 void                      * payload,
                                 size_t                      bytes,
                                 void                      * recv_func_parm,
                                 void                      * cookie);


  pami_result_t __PAMI_IBVSetDispatchFunc(pami_context_t    context,
					  size_t            dispatch,
					  pami_recvfunc_t   drecv_func,
					  void             *dr_func_parm,
					  size_t     *device_dispatch_id);
  
  pami_result_t __PAMI_IBVGetQPState (pami_context_t       context,
				      pami_ibv_qpstate_t  *qp_state);
  
  
  pami_result_t __PAMI_Invalidate_region (pami_context_t    ctxt,
					  void            * va,
					  size_t            bytes);

  pami_result_t __PAMI_Destroy_DCT (pami_context_t context);

#ifdef __cplusplus
};
#endif



#endif /* __default_pami_sys_post_h__ */
#endif /* __default_pami_sys_pre_h__ */
